// ThreadTest.cc
// -------------
// thread class test module
//                                               
// Design and Implementation by Bjoern Lemke               
//                                                         
// (C)opyright 2000 by Bjoern Lemke                        
//
//
///////////////////////////////////////////////////////////////////////////////

#include <unistd.h>
#include <pthread.h>

#include "Exception.h"
#include "Thread.h"
#include "ThreadLock.h"
#include "Chain.h"

#include "SigHandler.h"
#include "CegoFileHandler.h"
#include "CegoDatabaseManager.h"
#include "CegoLockHandler.h"


#define MAXTHREAD 100

long numPageAlloc;
long numThread;

static ThreadLock _lock;

class FHThread : public Thread
{
    
public:

    FHThread(CegoFileHandler *pFH, CegoDatabaseManager *pDBMng) : Thread()  {
	
	_pFH = pFH;
	_pLH = new CegoLockHandler(pDBMng);
	      
    }
    
    void* job(void* arg)  {
	
	try {

	    int n = *(int*)arg;      

	    long i=0;
	    while (i < numPageAlloc)
	    {

		int fileId;
		int pageId;
		unsigned* fbm;
		int fbmSize;

		// _lock.writeLock(100000);
		
		if ( i % 10000 == 0 )
		{
		    Chain msg = Chain("Thread ") + Chain(getTid()) + Chain(" : Allocating page ") + Chain(i);
		    cout << msg << endl;
		}

		_pFH->allocatePage(1, CegoFileHandler::DATAFILE,
				   fileId, 
				   pageId, 
				   _pLH, 
				   fbm, fbmSize);

		if ( i % 10000 == 0 )
		{
		    Chain msg = Chain("Thread ") + Chain(getTid()) + Chain(" : Page ") + Chain(fileId) + Chain("/") + Chain(pageId); 
		    cout << msg << endl;
		}	
		
		// cout << "Thread " << getTid() << " : Releasing page ... " << endl;
		_pFH->releasePage(fileId,
				  pageId, 
				  _pLH,
				  fbm, fbmSize);

		if ( i % 10000 == 0 )
		{
		    Chain msg = Chain("Thread ") + Chain(getTid()) + Chain(" : Released page ") + Chain(i);
		    cout << msg << endl;
		}


		// _lock.unlock();

		i++;
	    }
	    cout << "Thread " << getTid() << " : Finished " << endl;
	}
	catch (Exception e)
	{
	    e.print();
	    cout << "thread failed" << endl;
	}
	
    } 

private:
    

    CegoFileHandler *_pFH;
    CegoLockHandler *_pLH;
    
};

int main(int argc, char **argv)
{
  
    try {    


	numThread=atol(argv[1]);
	numPageAlloc=atol(argv[2]);

	cout << "NumThread=" << numThread << endl;
	cout << "NumPageAlloc=" << numPageAlloc << endl;
	
	CegoFileHandler *pFH = new CegoFileHandler("fhtest.log", 16384);
	CegoDatabaseManager *pDBMng = new CegoDatabaseManager("fh.xml", "fhtest.log");	
	pDBMng->xml2Doc();

	CegoLockHandler *pLH = new CegoLockHandler(pDBMng);
	pLH->initLocks();
	
	_lock.init();

	cout << "Init datafile .." << endl;
	pFH->initDataFile(1, "fhtest1.sys", 1, 2000, CegoFileHandler::SYSTEMFILE);
	pFH->initDataFile(1, "fhtest1.dta", 2, 2000, CegoFileHandler::DATAFILE);
	pFH->initDataFile(1, "fhtest2.dta", 3, 2000, CegoFileHandler::DATAFILE);
	pFH->initDataFile(1, "fhtest3.dta", 4, 2000, CegoFileHandler::DATAFILE);
	cout << "Reg datafile .." << endl;
	pFH->regDataFile(1, "fhtest1.sys", 1, pLH);
	pFH->regDataFile(1, "fhtest1.dta", 2, pLH);
	pFH->regDataFile(1, "fhtest2.dta", 3, pLH);
	pFH->regDataFile(1, "fhtest3.dta", 4, pLH);


	
	// _lock.writeLock(100000);
	
	cout << "Preallocating pages ..." << endl;
	for ( int i=0; i< 200; i++)
	{
	    int fileId;
	    int pageId;
	    unsigned* fbm;
	    int fbmSize;
	
	    pFH->allocatePage(1, CegoFileHandler::DATAFILE,
			       fileId, 
			       pageId, 
			       pLH, 
			       fbm, fbmSize);

	}
	cout << "---" << endl;
	FHThread* pFHThread[MAXTHREAD];
	
	int i, num[MAXTHREAD];
	long id[MAXTHREAD];

	for (i=0; i<numThread; i++) 
	{
	    
	    pFHThread[i] = new FHThread(pFH, pDBMng);
	    
	    num[i]=i;
	    
	    cout << "Starting thread " << id[i] << endl;

	    id[i] = pFHThread[i]->start(&num[i]);
	    
	    cout << "Thread " << id[i] << " started" << endl;       
	    // sleep(1);
	    
	}
	

	sleep(3);

	// myThread.V();
	for (i=0; i<numThread; i++) 
	    pFHThread[i]->join(id[i]);
	
	
	cout << "ok" << endl;       	
	
    }
    catch (Exception e)
    {
	e.print();
	cout << "failed" << endl;
    }
    
}

  
